%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%															     %
% autotalk.lgt - Automatas and Logtalk                           %
%															     %
%  - Paul Andrew Crocker										 %
%	 (Depart. de Informatica, Univ. Beira Interior, Covilhã)	 %
%															     %
% AutoTalk, version 1.0                                          %
% 1/Fev/2009												     %
%															     %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Object that represents a Finite Automata
% 

:- object(classFa,
	specializes(logtalk)).

	:- info([
		version is 1.0,
		author is 'Paul Crocker',
		date is 2008/12/01,
		comment is 'Object that Represents a Finite Automata.']).

	:- private(estadoInicial/1).
	:- private(estadoFinal/1).
	:- private(delta/3).
	:- private(epsilon/2).
	:- private(reconheceDfa/3).
	:- private(reconheceNfa/3).

	:- public(aceita/3).
	:- public(aceita/2).
	:- public(aceita/1).
	
	
	%DFA
	%
	aceitaDfa(Palavra, [InicialEstado|R] ):-
		::estadoInicial(InicialEstado),
		reconheceDfa(Palavra, InicialEstado, R).	
		
	reconheceDfa([], Estado, []):- 
		::estadoFinal(Estado), !.
		
	reconheceDfa([C|Resto], Estado, [ProximoEstado|R] ):- 
		::delta(Estado, C, ProximoEstado ), !,
		reconheceDfa(Resto, ProximoEstado, R ).
	%
	%END DFA	
		
	%NFA
	%
	aceitaNfa(Palavra, [InicialEstado|R]  ):-
		::estadoInicial(InicialEstado),
		reconheceNfa(Palavra, InicialEstado, R).
	
	reconheceNfa([],Estado, []):- 
		::estadoFinal(Estado).
	
	reconheceNfa([C|Resto], Estado, [ProximoEstado|R] ):- 
		::delta(Estado, C, ProximoEstado),
		reconheceNfa(Resto, ProximoEstado, R).
		
	reconheceNfa(Palavra, Estado, [ProximoEstado|R]):- 
		::epsilon(Estado,ProximoEstado),
		reconheceNfa(Palavra, ProximoEstado, R).
	%
	%END NFA	
		
	%Public FA
	
	aceita(dfa, Palavra ):- 	  aceitaDfa(Palavra, _).
	aceita(nfa, Palavra ):-       aceitaNfa(Palavra, _).
	
	aceita(Palavra,Trace):-
		::isaDfa,
		write('It is a Dfa.'),!,
   		aceitaDfa(Palavra, Trace).
 	
 	aceita(Palavra,Trace):-
		write('It is a Nfa.'),
    	aceitaNfa(Palavra, Trace).
	
	aceita(dfa, Palavra, Trace)	:- aceitaDfa(Palavra, Trace).
	aceita(nfa, Palavra, Trace)	:- aceitaNfa(Palavra, Trace).
	aceita(Palavra)				 :- aceita(Palavra,_).
	
	
	%End Public FA
	
	:- public(alphabet/1).
	alphabet(L):-
		::findall( X, delta(_,X,_), Y),
		removedups(Y,L).
	
	:- public(listInicial/1).
	listInicial(L):-
		::findall( X, estadoInicial(X), L).
	
	:- public(listFinal/1).
	listFinal(L):-
		::findall( X, estadoFinal(X), L).
		
	:- public(listStates/1).
	listStates(L):-
		::findall(X, delta(X,_,_),L1),
		::findall(Y, delta(_,_,Y),L2),
		list::append(L1,L2,L3),
		removedups(L3,L).
		
	:- public(isaDfa/0).
	% need to check that each state has only one transition for each member of the alphabet
	% usually check that there is only on start state
	% usually check that there is at least one final state
	
	isaDfa:-
		%check only one initial state
		alphabet(A1),
		list::sort(A1,A),		%sorting the list now is handy!
		listStates(S),
		check(S,A).
		
	check([], _ ).
	check([State|States], Alphabet):-
		::findall(X, delta(State,X,_), L),
		list::same_length(L,Alphabet),  %check exactly one transition per memeber of alphabet		
		list::sort(L,Alphabet),			%will use fact that sort removes duplicates, 
										%should be then equal to alphabet
		check(States, Alphabet).
		
	%convert automata to .jff xml file for jflap
	:- public(toJflap/1).
	toJflap(File):-
	    open(File, write, Stream),
		write(Stream, '<?xml version=\"1.0\"?>\n'),
		write(Stream, '<!--Created with Logtalk-->\n'),
		write(Stream, '<structure>\n'),
		write(Stream, '<type>fa</type>\n'),
		write(Stream, '<automaton>\n'),
		::listStates(S),
		::listInicial(I),
		::listFinal(F),
		wStates(Stream, S,I,F,0),
		wTrans(Stream, S),
		write(Stream, '</automaton>\n'),
		write(Stream, '</structure>\n').
	
	:- private(wStates/5).
	wStates(_, [],_,_,_).
	wStates(Stream, [H|T],I,F,Id) :-
		write(Stream, '<state id=\"'),
		write(Stream, Id),
		write(Stream, '" name=\"'),
		write(Stream, H),
		write(Stream, '\">\n'),
		write(Stream, '<label>'),     
		write(Stream, H) ,
		write(Stream, '</label>\n'),
		( list::member(H,I) -> write(Stream, '<initial />\n'); true),
		( list::member(H,F) -> write(Stream, '<final />\n'); true  ),
		write(Stream, '</state>\n'),
		ID is Id+1,
		wStates(Stream, T,I,F,ID).
	
	:- private(wTrans/2).
	wTrans(Stream, S) :- 
		forall( ::delta(F,A,T), wT(Stream, F,A,T,S) ).
				
	:- private(wT/5).
	wT(Stream, F,A,T,S):-
		write(Stream, '<transition>\n'),
		lookup(F,S,N1),
		write(Stream, '<from>'),write(Stream, N1),	write(Stream, '</from>'),
		lookup(T,S,N2),
		write(Stream, '<to>'),  write(Stream, N2),	write(Stream, '</to>')  ,
		write(Stream, '<read>'),write(Stream, A),	write(Stream, '</read>\n'),
		write(Stream, '</transition>\n').
	
	% Subtracts the elements of one list from another .. not used
	:- public(subtract/3).
	subtract([],_,[]).
	subtract([H|Tail], L2, L) :- list::member(H,L2), !,subtract(Tail,L2,L). 
	subtract([H|Tail], L2, [H|L]) :- 
		!,subtract(Tail,L2,L).	
		
	%predicate to lookup an element of a list and return its position in th elist -1 on fail
	:-private(lookup/3).
	:-private(d_lookup/4).
	lookup(E,L,N) :- d_lookup(E, L, 0, N).
	d_lookup(_,[],_, N) :- N is (-1).
	d_lookup(E,[E|_],N,N).
	d_lookup(E,[_|T],ID,N):- ID1 is ID+1,  d_lookup(E, T, ID1, N).
	
	:- private(removedups/2).
	% will use the standard prolog function member
	% hence will need to escape out of logtalk with the {} brackets
	removedups([],[]).
	removedups([X|Rest], Result) :- 
		{member(X,Rest)}, removedups(Rest,Result), !.
	removedups([X|Rest], [X|Rest1]) :- 
		not({member(X,Rest)}), removedups(Rest,Rest1),!.	

		
:- end_object.

